package cn.iocoder.yudao.module.bpm.controller.admin.task;

import cn.iocoder.yudao.framework.common.pojo.CommonResult;
import cn.iocoder.yudao.framework.common.pojo.PageResult;
import cn.iocoder.yudao.framework.common.util.json.JsonUtils;
import cn.iocoder.yudao.framework.idempotent.core.annotation.Idempotent;
import cn.iocoder.yudao.framework.security.core.LoginUser;
import cn.iocoder.yudao.framework.security.core.util.SecurityFrameworkUtils;
import cn.iocoder.yudao.module.bpm.controller.admin.definition.vo.copy.ProcessCopyPageReqVO;
import cn.iocoder.yudao.module.bpm.controller.admin.definition.vo.copy.ProcessCopyRespVO;
import cn.iocoder.yudao.module.bpm.controller.admin.task.vo.task.*;
import cn.iocoder.yudao.module.bpm.service.task.BpmProcessInstanceService;
import cn.iocoder.yudao.module.bpm.service.task.BpmTaskService;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiImplicitParam;
import io.swagger.annotations.ApiOperation;
import lombok.extern.slf4j.Slf4j;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.*;

import javax.annotation.Resource;
import javax.validation.Valid;
import java.util.List;

import static cn.iocoder.yudao.framework.common.pojo.CommonResult.success;
import static cn.iocoder.yudao.framework.web.core.util.WebFrameworkUtils.getLoginUserId;

/**
 * @author Administrator
 */
@Api(tags = "管理后台 - 流程任务实例")
@RestController
@RequestMapping("/bpm/task")
@Validated
@Slf4j
public class BpmTaskController {

    @Resource
    private BpmTaskService taskService;

    @Resource
    private BpmProcessInstanceService processInstanceService;

    @GetMapping("todo-page")
    @ApiOperation("获取 Todo 待办任务分页")
    @PreAuthorize("@ss.hasPermission('bpm:task:query')")
    public CommonResult<PageResult<BpmTaskTodoPageItemRespVO>> getTodoTaskPage(@Valid BpmTaskTodoPageReqVO pageVO) {
        return success(taskService.getTodoTaskPage(getLoginUserId(), pageVO));
    }

    @GetMapping("todo-count")
    @ApiOperation("获取 Todo 待办任务总数")
    public CommonResult<Long> getTodoTaskCount() {
        return success(taskService.getTodoTaskCount(getLoginUserId()));
    }

    @GetMapping("done-page")
    @ApiOperation("获取 Done 已办任务分页")
    @PreAuthorize("@ss.hasPermission('bpm:task:query')")
    public CommonResult<PageResult<BpmTaskDonePageItemRespVO>> getDoneTaskPage(@Valid BpmTaskDonePageReqVO pageVO) {
        return success(taskService.getDoneTaskPage(getLoginUserId(), pageVO));
    }

    @GetMapping("copy-page")
    @ApiOperation("获取抄送任务分页")
    @PreAuthorize("@ss.hasPermission('bpm:task:query')")
    public CommonResult<PageResult<ProcessCopyRespVO>> getCopyTaskPage(@Valid ProcessCopyPageReqVO pageVO) {
        return success(taskService.getCopyTaskPage(getLoginUserId(), pageVO));
    }

    @GetMapping("/list-by-process-instance-id")
    @ApiOperation(value = "获得指定流程实例的任务列表", notes = "包括完成的、未完成的")
    @ApiImplicitParam(name = "processInstanceId", value = "流程实例的编号", required = true, dataTypeClass = String.class)
    @PreAuthorize("@ss.hasPermission('bpm:task:query')")
    public CommonResult<List<BpmTaskRespVO>> getTaskListByProcessInstanceId(
            @RequestParam("processInstanceId") String processInstanceId) {
        return success(taskService.getTaskListByProcessInstanceId(processInstanceId));
    }

    @PutMapping("/approve")
    @ApiOperation("通过任务")
    @Idempotent(timeout = 5)
    @PreAuthorize("@ss.hasPermission('bpm:task:update')")
    public CommonResult<Boolean> approveTask(@Valid @RequestBody BpmTaskApproveReqVO reqVO) {
        //String jsonStr = JsonUtils.toJsonString(reqVO);
        //JsonUtils.SaveLog("参数为："+jsonStr);
        taskService.approveTask(getLoginUserId(), reqVO);
        return success(true);
    }

    /*
    lanbm 2024-05-20 add
     */
    @PutMapping("/approve2")
    @ApiOperation("批量审批通过任务")
    @Idempotent(timeout = 20)
    @PreAuthorize("@ss.hasPermission('bpm:task:update')")
    public CommonResult<Boolean> approveTask2(@Valid @RequestBody List<BpmTaskApproveReqVO> BpmList) {
        //先校验流程是否同一类，然后批量审批
        //String jsonStr = JsonUtils.toJsonString(BpmList);
        //JsonUtils.SaveLog("参数为："+jsonStr);
        long id=  getLoginUserId();
        for (BpmTaskApproveReqVO ReqVO : BpmList) {
            List<BpmTaskRespVO>taskList=
                    taskService.getTaskListByProcessInstanceId(ReqVO.getId());
            if(taskList.size()>0) {
                String sid=taskList.get(0).getId();
                ReqVO.setId(sid);
                //String jsonStr = JsonUtils.toJsonString(ReqVO);
                //JsonUtils.SaveLog("参数为："+jsonStr);
                taskService.approveTask(id, ReqVO);
            }
        }
        return success(true);
    }

    /*
     批量审批不通过 lanbm 2024-05-20 add
     */
    @PutMapping("/reject2")
    @ApiOperation("不通过任务")
    @Idempotent(timeout = 20)
    @PreAuthorize("@ss.hasPermission('bpm:task:update')")
    public CommonResult<Boolean> rejectTask2(@Valid @RequestBody List<BpmTaskRejectReqVO> BpmList) {
        //先校验流程是否同一类，然后批量审批
        LoginUser l=SecurityFrameworkUtils.getLoginUser();
        for (BpmTaskRejectReqVO reqVO : BpmList
        ) {
            List<BpmTaskRespVO>taskList=
                    taskService.getTaskListByProcessInstanceId(reqVO.getId());
            if(taskList.size()>0) {
                String sid=taskList.get(0).getId();
                reqVO.setId(sid);
                taskService.rejectTask(l, reqVO);
            }
        }
        return success(true);
    }


    @PutMapping("/reject")
    @ApiOperation("不通过任务")
    @Idempotent(timeout = 5)
    @PreAuthorize("@ss.hasPermission('bpm:task:update')")
    public CommonResult<Boolean> rejectTask(@Valid @RequestBody
                                                        BpmTaskRejectReqVO reqVO) {
        taskService.rejectTask(SecurityFrameworkUtils.getLoginUser(), reqVO);
        return success(true);
    }

    @PutMapping("/update-assignee")
    @ApiOperation(value = "更新任务的负责人", notes = "用于【流程详情】的【转派】按钮")
    @PreAuthorize("@ss.hasPermission('bpm:task:update')")
    public CommonResult<Boolean> updateTaskAssignee(@Valid @RequestBody BpmTaskUpdateAssigneeReqVO reqVO) {
        taskService.updateTaskAssignee(getLoginUserId(), reqVO);
        return success(true);
    }
}
